# This file is part of py-neonUtilities.

# py-neonUtilities is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# py-neonUtilities is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with py-neonUtilities.  If not, see <https://www.gnu.org/licenses/>.

import glob
from . import neon
import os
from os.path import join
from pathlib import Path
import shutil
import zipfile
import re
from itertools import chain


class NeonObservational(neon.Neon):
    def __init__(self, dpID=None, site=None, dates=None, package="basic", token=None):
        # inherit functions from the parent Neon class from neon.py
        neon.Neon.__init__(
            self, dpID=dpID, site=site, dates=dates, package=package, token=token
        )
        self.stackedFiles = {}

    def stackByTable(self, root=None, clean=True):
        """
        Method to stack zip files by table

        stackByTable can be used as a class method after downloading files, or you can provide a single argument root
        which is the path to the directory containing zip files. If used as a class method,
        it will create a directory named the dpID of your file. After stacking, the csv files containing
        data will be placed in <dpID>/stackedFiles. The csv name will follow the pattern DESC_stacked.csv.
        You can pass clean=False to not delete expanded files after stacking.
        """

        if not root:
            root = join(os.getcwd(), self.rootname)
        else:
            root = join(os.getcwd(), root)
            self.zipfiles = glob.glob(join(root, "*.zip"))

        if len(self.zipfiles) == 0:
            print(
                "No download files found. Pass the download directory to this function or use the download() method."
            )
            return

        # set the root to the default if the user gave nothing.
        self.root = (
            join(os.getcwd(), self.rootname) if not root else join(os.getcwd(), root)
        )

        # defaults to the equivalent in the R package for compatibility.
        self.stackedDir = join(os.getcwd(), root, "stackedFiles")

        self.resetDir(self.stackedDir)

        files = []
        self.zipfiles = sorted(self.zipfiles)
        # unzip all. sorted to make sure everything is in order.
        for fpath in self.zipfiles:
            with zipfile.ZipFile(fpath, "r") as f:
                Path(join(self.root, fpath[:-4])).mkdir(parents=True, exist_ok=True)
                f.extractall(join(self.root, fpath[:-4]))
                files.append([join(self.root, fpath[:-4], i) for i in f.namelist()])

        # siteDateRE = re.compile("\.[a-z]{3}_(.*)\.[0-9]{4}-[0-9]{2}\." + self.data["package"] + "(.*)\.csv")
        # siteAllRE = re.compile("\.[a-z]{3}_([a-z]*)\."+self.data["package"]+"(.*)\.csv")
        # expanded packages sometimes contain basic files.
        siteDateRE = re.compile(
            "\.[a-z]{3}_(.*)\.[0-9]{4}-[0-9]{2}\." + "[a-z]*" + "(.*)\.csv"
        )
        siteAllRE = re.compile("\.[a-z]{3}_([a-zA-Z]*)\.[a-z]*\.(.*)\.csv")
        labRE = re.compile("NEON\.(.*)\.[a-z]{3}_([a-zA-Z]*)\.csv")
        self.labRE = labRE

        self.siteDateFiles = [[i for i in j if siteDateRE.search(i)] for j in files]
        self.siteAllFiles = [[i for i in j if siteAllRE.search(i)] for j in files]
        self.labFiles = [
            [i for i in j if labRE.match(os.path.basename(i))] for j in files
        ]

        self.stackedFiles = {}

        flat = set(
            [self.extractName(i) for i in list(chain.from_iterable(self.siteDateFiles))]
        )
        # key functions
        self.stack_site_date(self.siteDateFiles, flat)
        self.stack_site_all()
        self.stack_lab()

        if clean:
            # inherited
            self.cleandir(self.root)

    def instances(self, name, files):
        """Helper function for *-all stacking. Finds tables
        with the same name from each site.
        """

        # requires files to be presorted.
        inst = {i: None for i in self.data["site"]}
        for i in files:
            for j in i:
                if name in j:
                    inst[self.extractLoc(j)] = j
                    continue
        return inst

    def stack_site_all(self):
        """stack site-all"""

        # TODO file may not be in the latest version.
        if len(self.siteAllFiles) == 0:
            return
        for i in self.siteAllFiles[0]:
            name = self.extractName(i)
            toStack = self.instances(name, self.siteAllFiles)
            filename = join(self.stackedDir, name + "_stacked.csv")
            outf = neon.CSVwriter(filename)
            for j in toStack:
                if not toStack[j]:
                    continue
                outf.append(join(self.root, toStack[j]))

            outf.close()
            self.stackedFiles[name] = filename

    def stack_lab(self):
        if len(self.labFiles) == 0:
            return
        for i in self.labFiles[0]:
            name = self.extractName(i)
            if self.is_lab_all(name, self.extractLoc(i), self.labFiles):
                self.stack_lab_all(name)
            else:
                self.stack_lab_cur(name)

    def stack_lab_all(self, name):
        toStack = self.instances(name, self.labFiles)
        filename = join(self.stackedDir, name + "_stacked.csv")
        outf = neon.CSVwriter(filename)
        for j in toStack:
            if not toStack[j]:
                continue
            outf.append(join(self.root, toStack[j]))
        outf.close()
        self.stackedFiles[name] = filename

    def stack_lab_cur(self, name):
        flat = set([i for i in list(chain.from_iterable(self.labFiles)) if name in i])
        filename = join(self.stackedDir, name + "_stacked")
        out = neon.CSVwriter(filename)
        for f in flat:
            out.append(f)
        out.close()

    def is_lab_all(self, name, lab, files):
        # print([i for i in list(chain.from_iterable(files)) if name in i])
        hashes = [
            self.hashf(i)
            for i in list(chain.from_iterable(files))
            if name in i and lab in i
        ]
        # if len(hashes)!=len(set(hashes)):
        if 1 == len(set(hashes)):
            return True
        return False

    def extractName(self, s):
        s = os.path.basename(s)
        match = self.nameRE.search(s)
        if not match:
            return None
        matchstr = str(match.group(0))
        if matchstr.count(".") == 3:
            return matchstr.split(".")[-2]
        if matchstr.count(".") == 4:
            return s[match.start() + 8 : match.end() - 7]
        if matchstr.count(".") == 5:
            return s[match.start() + 8 : match.end() - 15]

    def extractLoc(self, s):
        s = os.path.basename(s)
        if self.labRE.match(s):
            return s.split(".")[1]
        return os.path.basename(s).split(".")[2]




# This file is part of py-neonUtilities.

# py-neonUtilities is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# py-neonUtilities is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with py-neonUtilities.  If not, see <https://www.gnu.org/licenses/>.

from . import neon
import re
import os
from os.path import join
import shutil
from itertools import chain
import zipfile
from pathlib import Path


class NeonInstrumental(neon.Neon):
    def __init__(
        self, dpID=None, site=None, dates=None, avg=None, package="basic", token=None
    ):
        # inherit functions from the parent Neon class from neon.py

        neon.Neon.__init__(self, dpID, site, dates, avg, package, token)
        self.stackedFiles = {}

    def unzipAll(self, root):
        folders = []
        self.zipfiles = sorted(self.zipfiles)
        # unzip all. sorted to make sure everything is in order.
        for fpath in self.zipfiles:
            try:
                with zipfile.ZipFile(fpath, "r") as f:
                    Path(join(self.root, fpath[:-4])).mkdir(parents=True, exist_ok=True)
                    f.extractall(join(self.root, fpath[:-4]))
                    folders.append(join(self.root, fpath[:-4]))
            except:
                print(fpath,"bad zip file. Skipping.")
        return folders

    def stackByTable(self, root=None, clean=True, bySite=False):

        if root is None:
            self.root = join(os.getcwd(), self.rootname)
        else:
            self.folders = []
            self.root = join(os.getcwd(), root)
            folderfiles = os.listdir(self.root)
            for i in folderfiles:
                if os.path.isdir(i) and "stackedFiles" not in i:
                    self.folders.append(i)
                elif i.endswith(".zip"):
                    self.zipfiles.append(join(self.root,i))

        if len(self.folders) == 0 and len(self.zipfiles) == 0:
            print(
                "No files stacked. Use download() or pass the folder path to stackByTable."
            )
            return
        elif len(self.folders) == 0 and len(self.zipfiles) > 0:
            self.folders = self.unzipAll(self.root)

        self.stackedDir = join(os.getcwd(), self.root, "stackedFiles")
        self.resetDir(self.stackedDir)

        self.folders = sorted(self.folders)
        self.files = []

        for i in self.folders:
            self.files.append([join(i, j) for j in os.listdir(join(self.root, i))])

        flat = set(
            [self.extractISname(i) for i in list(chain.from_iterable(self.files))]
        )

        if bySite:
            for site in self.data["site"]:
                self.stack_site_date(self.files, flat, site=site)
        else:
            self.stack_site_date(self.files, flat)

        if clean:
            self.cleandir(self.root)

    def extractISname(self, s):
        s = os.path.basename(s)
        match = self.isre.search(s)
        if not match:
            return None
        matchstr = str(match.group(0))
        if self.data["avg"] is None:
            return matchstr.split(".")[-5]
        else:
            return matchstr.split(".")[-1]



# This file is part of py-neonUtilities.

# py-neonUtilities is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# py-neonUtilities is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with py-neonUtilities.  If not, see <https://www.gnu.org/licenses/>.

# TODO make sure that you can actually only provide dates and not site.

import json
import os
from os.path import join
import re
import time
import requests as req
import urllib
import glob
import shutil
import zlib
import math


class Neon:
    """Parent class for all Neon datatypes"""

    def __init__(
        self, dpID=None, site=None, dates=None, avg=None, package="basic", token=None
    ):
        if package != "basic" and package != "expanded":
            print("package must be either basic or expanded. defaulting to basic.")
            package = "basic"

        self.data = {
            "dpID": dpID,
            "site": site,
            "dates": dates,
            "avg": avg,
            "package": package,
            "token": token,
        }

        self.baseurl = "https://data.neonscience.org/api/v0/data/"
        self.zipre = re.compile(package + "\.(.*)\.zip")
        self.nameRE = re.compile(
            "NEON\.(.*)\.[a-z]{3}_([a-zA-Z]*)\.csv|[0-9]{3}\.(.*)\.([0-9]{4}-[0-9]{2}|[a-z]*)\."
        )

        if avg != None:
            if type(avg) == int:
                avg = str(avg)
                self.isre = re.compile(
                    "[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.(.*)_" + avg + "min"
                )
            else:
                avg = str(avg)
                self.isre = re.compile(
                    "[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.(.*)_" + avg
                )
        else:
            self.isre = re.compile("[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.[0-9]{3}\.(.*)_(.*)")

        self.filere = re.compile(".csv")
        self.packagere = re.compile(package)
        self.folders = []
        self.zipfiles = []

    def makeIfNotExists(self, name):
        if not os.path.exists(name):
            os.makedirs(name)
            return True
        return False

    def download(self):
        """Class method to download zip files. Overridden."""
        self.rootname = self.data["dpID"]

        # create dataproduct directory if it does not exist
        # TODO allow for multiple dataproducts to be downloaded at once and stacked.
        if self.makeIfNotExists(self.rootname):
            print(f"[created root folder {self.rootname}]")

        # find all the idxurls of month-chunks
        self.idxurls = self.constructIdxUrls()
        print(f"{len(self.idxurls)} file(s) in total")
        self.zipfiles = []

        res = False
        for n, idxurl in enumerate(self.idxurls):
            self.currentlyDl = n
            if self.data["avg"] != None:
                res = self.downloadFiles(idxurl, usere=self.isre)
                # TODO hashchecking
                # TODO sizechecking
            else:
                res = self.downloadZips(idxurl)
        if res:
            print("Done downloading.")

    def makeReq(self, url):
        """wrapper function to allow for API token requests"""
        if self.data["token"]:
            return req.get(url, headers={"X-API-TOKEN": self.data["token"]})
        else:
            return req.get(url)

    def getReq(self, idxurl):
        """catch ratelimit exceeded and wait"""
        req = self.makeReq(idxurl)
        while req.headers["X-RateLimit-Remaining"] == 1:
            print("Rate limit exceeded. Consider using an api token. Pausing...")
            time.sleep(req.headers["RetryAfter"])
            print("Retrying")
            req = self.makeReq(idxurl)

        return json.loads(req.text)["data"]["files"]

    def readable(self, size):
        if size == 0:
            return "0"
        exts = ("B", "KB", "MB", "GB")
        ext = int(math.floor(math.log(size, 1024)))
        val = round(size / math.pow(1024, ext), 2)
        return f"{val} {exts[ext]}"

    def downloadZips(self, idxurl):
        """privately used function to download an individual zip file from url"""
        index = self.getReq(idxurl)

        zipidx = None
        for i in range(len(index)):
            match = self.zipre.search(index[i]["name"])
            if match:
                zipidx = i
                break
        if zipidx != None:
            size = int(index[zipidx]["size"])
            print(
                f"Downloading chunk {self.currentlyDl+1}. Size: {self.readable(size)}"
            )
            #TODO move this check before the request.
            if os.path.exists(os.path.join(self.rootname,index[zipidx]['name'])):
                print("already downloaded")
                return True
            try:
                urllib.request.urlretrieve(
                    index[zipidx]["url"], os.path.join(self.rootname, index[zipidx]["name"])
                )
            except:
                print("Download failed, continuing...")

            self.zipfiles.append(
                os.path.join(os.getcwd(), self.rootname, index[zipidx]["name"])
            )
            print(f"Completed zip {self.currentlyDl+1}")
        else:
            print(
                f"Zip file missing. This may be because this data chunk {idxurl} does not exist."
            )
            return False
        return True

    def downloadFiles(self, idxurl, usere=None):
        """Downloads files instead of zips"""
        index = self.getReq(idxurl)
        foldername = None
        for i in index:
            if self.zipre.search(i["name"]):
                foldername = i["name"][:-4]
                break

        if not foldername:
            print(
                f"File missing. This may be because data chunk {idxurl} does not exist."
            )
            return False
        self.makeIfNotExists(os.path.join(self.rootname, foldername))

        if not usere:
            usere = self.filere

        print(f"Downloading chunk {self.currentlyDl+1}.")
        for i in index:
            if usere.search(i["name"]) and self.packagere.search(i["name"]):
                urllib.request.urlretrieve(
                    i["url"], os.path.join(self.rootname, foldername, i["name"])
                )
        self.folders.append(foldername)
        print(f"Completed chunk {self.currentlyDl+1}")
        return True

    def mkdt(self, y, m):
        if m < 10:
            m = "0" + str(m)
        return str(y) + "-" + str(m)

    def getRangeDates(self, pair):
        if type(pair) == str:
            return [pair]
        sdate, edate = pair
        dates = []
        sy, sm = int(sdate.split("-")[0]), int(sdate.split("-")[1])
        ey, em = int(edate.split("-")[0]), int(edate.split("-")[1])
        if ey == sy:
            dates.extend([self.mkdt(sy, i) for i in range(sm, em + 1)])
        else:
            dates.extend([self.mkdt(sy, i) for i in range(sm, 13)])
            for i in range(sy + 1, ey):
                dates.extend([self.mkdt(i, j) for j in range(1, 13)])
            dates.extend([self.mkdt(ey, i) for i in range(1, em + 1)])
        return dates

    def basicUrl(self, dpid, site, date):
        return self.baseurl + dpid + "/" + site + "/" + date

    def constructIdxUrls(self):
        # if there is only one site as a string, turn it into a single element array.
        if type(self.data["site"]) == str:
            self.data["site"] = [self.data["site"]]

        urls = []
        if type(self.data["dates"]) == dict:
            for i in self.data["dates"]:
                dates = []
                for date in self.data["dates"][i]:
                    dates.extend(self.getRangeDates(date))
                urls.extend(
                    [self.basicUrl(self.data["dpID"], i, date) for date in dates]
                )
        else:
            dates = []
            for date in self.data["dates"]:
                dates.extend(self.getRangeDates(date))

            for site in self.data["site"]:
                urls.extend(
                    [self.basicUrl(self.data["dpID"], site, date) for date in dates]
                )

        return urls

    def cleandir(self, direc):
        toRemove = [
            i
            for i in glob.glob(os.path.join(direc, "*"))
            if not self.zipre.search(i) and "stackedFiles" not in i
        ]
        for i in toRemove:
            if os.path.isdir(i):
                shutil.rmtree(i)
            else:
                os.remove(i)

    def resetDir(self, direc):
        if os.path.exists(direc):
            shutil.rmtree(direc)
        os.makedirs(direc)

    def hashf(self, filename):
        with open(filename, "rb") as fh:
            hash = 0
            while True:
                s = fh.read(65536)
                if not s:
                    break
                hash = zlib.crc32(s, hash)
        return "%08X" % (hash)

    def to_pandas(self):
        """Converts a stacked dataset into a dictionary of pandas DataFrames"""
        if len(self.stackedFiles) == 0:
            print("No files stacked")
            return
        import pandas as pd

        dfs = {}
        for i in self.stackedFiles:
            if type(self.stackedFiles[i])==str:
                dfs[i] = pd.read_csv(self.stackedFiles[i])
            else:
                for n,name in enumerate(self.stackedFiles[i]):
                    if i in dfs:
                        dfs[i][name] = filename
                    else:
                        dfs[i] = {name:pd.read_csv(self.stackedFiles[i][name])}

        return dfs

    def stack_site_date(self, files, flatnames, site=None):
        for name in flatnames:
            if not name:
                continue
            if site is None:
                filename = join(self.stackedDir, name + "_stacked.csv")
            else:
                filename = join(self.stackedDir, site+name+"_stacked.csv")

            out = CSVwriter(filename)
            for other in range(len(files)):
                for i in files[other]:
                    if (name in i) and (site is None or site in i):
                        out.append(join(self.root, i))
                        break
            out.close()
            if site is None:
                self.stackedFiles[name] = filename
            else:
                if site in self.stackedFiles:
                    self.stackedFiles[site][name] = filename
                else:
                    self.stackedFiles[site] = {name:filename}


class CSVwriter:
    def __init__(self, outname):
        self.filename = outname
        self.out = open(self.filename, "a")
        self.first = True

    def append(self, inname):
        with open(inname) as otherf:
            if not self.first:
                otherf.__next__()
            for line in otherf:
                self.out.write(line)
            self.first = False

    def close(self):
        self.out.close()



# This file is part of py-neonUtilities.

# py-neonUtilities is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# py-neonUtilities is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with py-neonUtilities.  If not, see <https://www.gnu.org/licenses/>.

from importlib import reload
import neon
reload(neon)

class NeonAOP(neon.Neon):
    def __init__(self, dpID=None, site=None, year=None):
        #TODO error checking.
        pass



import setuptools
import os
with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
      name='py-neonUtilities',
      version='0.5.4',
      description='A native python wrapper for the NEON ecological database api',
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=setuptools.find_packages(),
      python_requires='>=3.6',
      author='esrh',
      author_email='esrh@netc.eu',
      keywords='NEON data ecology api',
      url='https://gitlab.com/esrh/py-neonutils')
